/*
*   File: findall.pro
*   Author: Arun Lakhotia
*   Date: November 87
*
*   A findAll interpreter. Performs action similar to findall/3 but
*   does it by interpreting the program and collecting solutions in its
*   argument, rather than the conventional method of executing the
*   goal under Prolog and collecting solutions by side-effect
*
*   POTENTIAL USE:
*     Transforming "generate-one" versions of programs to
*     "generate-all" versions by compiling with this interpreter
*
*   MODIFICATIONS
*    Feb 15, 1990
*      Removed the definitions of copy/2 and unify/2 from here.
*      The reified definitions of these predicates may be used for
*      specialization
*
*/

%%  findAll(+Term, +Goal, -AllSolutions)
%%    same as findall/3
%%    except: it is implemented as an interpreter
%%            operates on cut free 
%%    Limitation: Because it cannot interpret `!'-full (;-)
%%        programs, and it itself uses `!', findAll/3
%%        cannot work on itself. Thus findAll/3 cannot be
%%        nested like findAll(X, findAll(....), Y)

findAll(Term, Goal, Bag) :-
   findAlld(Term, Goal, Bag-[]).

%%  findAlld(+Term, +Goal, -AllSolutions)
%%     differs from findAll/3 as AllSolutions returned is a
%%     difference list, Head-Tail

    % Any Term is a solution of true
findAlld(Term, true, [Term|Bag]-Bag) :-!.
findAlld(Term, (Goal,Conjunction), Bag-List) :- !,
      % Hard to explain, leave it for the reader to figure out
        findAlld((Term,Conjunction), Goal, Pairs-[]),
	findAll_conjunct_d(Pairs, Bag-List). 
findAlld(Term, Goal, SystemBag) :-
	sys(Goal),!,
	findAll_system_predicate(Term, Goal, SystemBag).
findAlld(Term, Goal, Bag-Terms) :- 
      % get the list of clauses whose heads match goal
        all_clauses(Goal, ClauseList),
      % solve bodies and get bag of resulting heads
        findAll_clauses_d(ClauseList, SolutionHeads-[]),!,
      % extract solutions for Term from list of solutions
	extract_solutions_d(Term, Goal, SolutionHeads, Bag-Terms). 
   % if no solution, return an empty bag 
findAlld(_Term, _Goal, Bag-Bag).

       %% findAll_conjunct_d(+TermConjunctions, -Solutions)
       %%   TermConjunctions is a list of (Term, Conjunction) pair
       %%   Conjunction is the sequence of goals remaining to be executed
       %%   I find this hard to explain further. Please refer to the handling
       %%   of conjunctive case above
     
        findAll_conjunct_d([], Bag-Bag).
        findAll_conjunct_d([(Term,Conjunction)|Pairs], List1-List2) :- 
             findAlld(Term, Conjunction, List1-List1d),
             findAll_conjunct_d(Pairs, List1d-List2).

       %% findAll_clauses_d(+ClauseList, -Solutions)
       %%  ClauseList is a list of (Head :- Body) where Head unifies
       %%  some goal. This is a list of all such clauses in the database
       %%  The Solutions returned is a list of Head"'s, where each Head"
       %%  constitute a solution due to executing some clause in the list

        findAll_clauses_d([], Bag-Bag).
        findAll_clauses_d([(Head :-Body)|Pairs], List1-List2) :- 
             findAlld(Head, Body, List1-List1d),
             findAll_clauses_d(Pairs, List1d-List2).

       %% findAll_system_predicate(+Term, +Goal, -Solutions)
       %%   finds all solution for a system Goal.
       %%   uses systems findall/3 for this purpose
       %%   Note: the solution returned is not a difference list
       %%   and thus should be converted to one

       findAll_system_predicate(Term, Goal, Bag-Tail) :-
            findall(Term, Goal, SystemBag),
            make_diff_list(SystemBag, Bag-Tail).

extract_solutions_d(_, _,  [], Tail-Tail).
extract_solutions_d(Term, Goal, [Solution|Rest], [R|H]-T) :- 
        extract_solution(Term, Goal, Solution, R), !,
        extract_solutions_d(Term, Goal, Rest, H-T). 
extract_solutions_d(Term, Goal, [_Solution|Rest], H-T) :- 
       % not extract_solution(Term, Goal, _Solution, R), 
        extract_solutions_d(Term, Goal, Rest, H-T). 

extract_solution(Term, Goal, Solved_Goal, CopyTerm) :- 
        copy((Term,Goal), (CopyTerm,CopyGoal)),
	unify(CopyGoal, Solved_Goal).


