
/*
Here is a Prolog meta-interpreter that reifies unification,
in the sense that it operates on ground goals and clauses.
It is the first interprter of this kind that I write, and I have found
it surprisingly simple.   The key to its simplicity is the
variable-value dictionary, that uses unification of meta-level logical
variables to represent some of the variable-value bindings at the
object level.

      Udi

Sample runs are:

| ?- solve((member(var(x),[a,b,c]),member(var(x),[d,b,e])),A).

A = member(b,[a,b,c]),member(b,[d,b,e]) ;

no


| ?- solve((member(var(x),[a,b,var(y)]),member(var(x),[c,b,e])),A).

A = member(b,[a,b,var(y)]),member(b,[c,b,e]) ;

A = member(c,[a,b,c]),member(c,[c,b,e]) ;

A = member(b,[a,b,b]),member(b,[c,b,e]) ;

A = member(e,[a,b,e]),member(e,[c,b,e]) ;

no


?- solve((member(var(x),[a,b,var(y)]),member(var(z),[c,b,e])),A).

A = member(a,[a,b,var(y)]),member(c,[c,b,e]) ;

A = member(a,[a,b,var(y)]),member(b,[c,b,e]) ;

A = member(a,[a,b,var(y)]),member(e,[c,b,e]) ;

A = member(b,[a,b,var(y)]),member(c,[c,b,e]) ;

A = member(b,[a,b,var(y)]),member(b,[c,b,e]) ;

A = member(b,[a,b,var(y)]),member(e,[c,b,e]) ;

A = member(var(7),[a,b,var(7)]),member(c,[c,b,e]) ;

A = member(var(7),[a,b,var(7)]),member(b,[c,b,e])


*/


solve(A,A1) :-
        solve(A,1,_,D),
        extract(A,D,A1).

solve(Goal, InN, OutN, D) :-
     Goal = true -> InN = OutN
     ; Goal = (A,B) ->
          solve(A, InN, MidN, D), 
          solve(B, MidN, OutN, D)
     ; clause(Goal1, Body, InN, MidN),
       unify(Goal, Goal1, D),
       solve(Body, MidN, OutN, D).

unify(X, Y, D) :-
   (var(X); var(Y)) -> X = Y
   ; X = var(N) ->
     associate(N, V, D),
     unify(V, Y, D)
   ; Y = var(N) ->
        associate(N,V,D),
        unify(X,V,D)
   ; X =[A|As], Y = [B|Bs]
     -> unify(A, B, D), unify(As, Bs, D)
   ; atomic(X)
     -> X = Y
   ; X =..[F|Xs], Y =..[F|Ys], unify(Xs, Ys, D).


extract(X, D, V1) :-
   X = var(N)
    -> associate(N, V, D),
       extract_value(N, V, D, V1)
    ; atomic(X) -> V1 = X
    ; X = [A|As], V1 = [V|Vs]
      -> extract(A, D, V),
         extract(As, D, Vs)
    ; X =..[F|Xs],
      extract(Xs, D, Xs1),
      V1 =..[F|Xs1].

extract_value(N, X, D, Y) :-
       X = var(N),
       Y = var(N)
       -> true
       ; extract(X, D, Y).



associate(Key,Value,[Key1-Value1|D]) :-
      Key = Key1,
      Value = Value1
      -> true
      ; associate(Key, Value, D).

/*
clause(member(X,[X|Xs]),true,N1,N3) :-
        X=var(N1), N2 is N1+1,
        Xs=var(N2), N3 is N2+1.
clause(member(X,[X1|Xs]),member(X,Xs),N1,N4) :-
        X=var(N1), N2 is N1+1,
        Xs=var(N2), N3 is N2+1,
        X1=var(N3), N4 is N3+1.


Note: A possible optimization is to represent  only
nonlocal variables (i.e. variables that are in structures or only in
the body) by ground terms, and represent local variables by themselves.
This is a sufficient to ensure that if the goal is ground
so are its subgoals.

Example: in the following Ys is local.

clause(rev([X|Xs],Ys),(rev(Xs,Zs), append(Zs,[X],Ys)),N1,N4) :-
        X=var(N1), N2 := N1+1,
        Xs=var(N2), N3 := N2+1,
        Zs=var(N3), N4 := N3+1.

*/

