

/*
Here is a Prolog meta-interpreter that reifies unification,
in the sense that it operates on ground goals and clauses.
It is the first interprter of this kind that I write, and I have found
it surprisingly simple.   The key to its simplicity is the
variable-value dictionary, that uses unification of meta-level logical
variables to represent some of the variable-value bindings at the
object level.

      Udi

Sample runs are:

| ?- solve((member(var(x),[a,b,c]),member(var(x),[d,b,e])),A).

A = member(b,[a,b,c]),member(b,[d,b,e]) ;

no


| ?- solve((member(var(x),[a,b,var(y)]),member(var(x),[c,b,e])),A).

A = member(b,[a,b,var(y)]),member(b,[c,b,e]) ;

A = member(c,[a,b,c]),member(c,[c,b,e]) ;

A = member(b,[a,b,b]),member(b,[c,b,e]) ;

A = member(e,[a,b,e]),member(e,[c,b,e]) ;

no


?- solve((member(var(x),[a,b,var(y)]),member(var(z),[c,b,e])),A).

A = member(a,[a,b,var(y)]),member(c,[c,b,e]) ;

A = member(a,[a,b,var(y)]),member(b,[c,b,e]) ;

A = member(a,[a,b,var(y)]),member(e,[c,b,e]) ;

A = member(b,[a,b,var(y)]),member(c,[c,b,e]) ;

A = member(b,[a,b,var(y)]),member(b,[c,b,e]) ;

A = member(b,[a,b,var(y)]),member(e,[c,b,e]) ;

A = member(var(7),[a,b,var(7)]),member(c,[c,b,e]) ;

A = member(var(7),[a,b,var(7)]),member(b,[c,b,e])


*/

solve(A,A1) :-
        solve(A,1,_,D),
        extract(A,D,A1).

solve(true,N,N,_) :- !.
solve((A,B),N1,N3,D) :-  !,
        solve(A,N1,N2,D), solve(B,N2,N3,D).
solve(A,N1,N3,D) :-
        clause(A1,B,N1,N2),
        unify(A,A1,D),
        solve(B,N2,N3,D).

unify(X,Y,_) :- ( var(X) ; var(Y) ), !, X=Y.
unify(var(N),Y,D) :- !,
        associate(N,V,D),
        unify(V,Y,D).
unify(X,var(N),D) :- !,
        associate(N,V,D),
        unify(X,V,D).
unify([X|Xs],[Y|Ys],D) :- !,
        unify(X,Y,D), unify(Xs,Ys,D).
unify(X,Y,_) :- atomic(X), !, X=Y.
unify(T1,T2,D) :-
        T1=..[F|Xs], T2=..[F|Ys], unify(Xs,Ys,D).

extract(var(N),D,V1) :-  !,
        associate(N,V,D),
        extract_value(N,V,D,V1).
extract(C,D,C) :- atomic(C), !.
extract([X|Xs],D,[X1|Xs1]) :- !,
        extract(X,D,X1), extract(Xs,D,Xs1).
extract(A,D,A1) :-
        A=..[F|Xs],
        extract(Xs,D,Xs1),
        A1=..[F|Xs1].

extract_value(N,var(N),D,var(N)) :- !.
extract_value(N,V,D,V1) :-
        extract(V,D,V1).

associate(Key,Value,[Key-Value|D]) :- !.
associate(Key,Value,[Key1-Value1|D]) :-
        associate(Key,Value,D).


/*

clause(member(X,[X|Xs]),true,N1,N3) :-
        X=var(N1), N2 is N1+1,
        Xs=var(N2), N3 is N2+1.
clause(member(X,[X1|Xs]),member(X,Xs),N1,N4) :-
        X=var(N1), N2 is N1+1,
        Xs=var(N2), N3 is N2+1,
        X1=var(N3), N4 is N3+1.


Note: A possible optimization is to represent  only
nonlocal variables (i.e. variables that are in structures or only in
the body) by ground terms, and represent local variables by themselves.
This is a sufficient to ensure that if the goal is ground
so are its subgoals.

Example: in the following Ys is local.

clause(rev([X|Xs],Ys),(rev(Xs,Zs), append(Zs,[X],Ys)),N1,N4) :-
        X=var(N1), N2 := N1+1,
        Xs=var(N2), N3 := N2+1,
        Zs=var(N3), N4 := N3+1.

*/

