/*  Copyright (C) 1989,90  Arun Lakhotia    */
/* File: mix.pro
*  Author: Arun Lakhotia
*  Released:  Thu Jul 5 16:37:16 CDT 1990
*   $Log:	mix.pro,v $
**   Revision 1.2  90/07/05  17:24:18  arun
**   New Release
**   
**   Revision 1.1  89/04/16  01:29:08  arun
**   Initial revision
**         */
/* CHANGES
*      February 18, 1990 by AL
*       Fix for problem with evaluating in virtual database.
*       Replaced is_interpreted/1 by is_user_defined/1
*/ 

/* 
*       The core of the partial evaluator houses here.  Created afresh
*       after phasing out old code that served almost 2 years.  
*
*       It takes care of ALL control structures and predicates of
*       Prolog.  Code for meta predicates and set predicates is in the
*       files set.pro and meta.pro
*
*   CUT-LESS Version
*     The "transparent" and "opaque" versions of various predicates
*     perform identically. They are maintained just to maintain
*     similarity between the With-Cut and CUT-LESS versions.
*
*/

/*
*    ENTRY POINTS
*      -   drive_mix(+GOAL, CLAUSES)
*      -   mix(+GOAL, -RESIDUE)
*
*     The following returns a "bag" of all the RESIDUES. 
*      -   mix_bag(+GOAL, -RESIDUES)
*
*    Entry point USED BY THE DRIVER FOR SPECIALIZING INTERPRETERS. The
*    driver performs one level of unfolding and passes the unfolded
*    CLAUSES.
*      -   mix_clauses(HEAD, CLAUSES, RESIDUE)
*/


/*%  drive_mix(+GOAL, -MIXEDCLAUSE)
%     A driver for mix for the specific purpose of specializing
%     interpreters. It performs one level of unfolding of the
%     interpreter GOAL. The other levels are suppressed by the
%     unfold rules
%     MIXEDCLAUSE is a clause from mixing GOAL. Only one clause is
%     returned at a time. 
*/

drive_mix(Goal, MixedClause) :-
    is_user_defined(Goal),
    all_clauses(Goal, Clauses),
    mix_clauses(Goal, Clauses, MixedBody),
    MixedClause = (Goal:-MixedBody).

/*%  mix(+GOALS, -RESIDUE)
%       partially evaluates GOALS and returns the suspended computation
%       as RESIDUE
%       Cuts are ignored.
*/

  mix(Var, Residue) :-
      var(Var),!,
      make_atomic_residue(call(Var), Residue).
  mix(true, Residue):-
      !, make_atomic_residue(true, Residue).
  mix((X,Y), MixXY) :-!,
       mix_and(X, Y, MixXY).
  mix(X, Residue) :-
      is_control_predicate(X), !,
      mix_meta(X, Residue).
  mix(X, Residue) :-
      is_set_predicate(X), !,
      mix_set(X, Residue).
  mix(Goal, Residue) :-
      mix_atomic(Goal, Residue).

/*%% CONJUNCTIVE GOALS  %%
*/

  mix_and(X, Y, MixXY) :-
      ( var(X)
      ; is_side_effect(X)
      ; is_extra_logical(X)
      ),!,
       mix(X, MixX),
       mix_bag(Y, MixY), 
       merge_ands(MixX, MixY, MixXY).
/*       % no need to cut choice points since X cannot have a !
*/
  mix_and(GoalA, GoalB, MixAB) :-
      mix(GoalA, MixA),
      mix_conj(MixA, GoalB, MixAB).

     mix_conj(MixX, Y, MixXY) :-
	  mix(Y, MixY),
	  merge_ands(MixX, MixY, MixXY).

/*%% ATOMIC GOALS  %%%
*/

  mix_atomic(!, Residue) :-!, make_cut_residue(!, Residue).
  mix_atomic(X, Residue) :-
      immediately_evaluable(X), !, 
      evaluate(X),
      make_atomic_residue(true, Residue).
  mix_atomic(Head, MixB) :-
      is_user_defined(Head),
      should_unfold(Head), !,
      all_clauses(Head, Clauses),
      mix_clauses(Head, Clauses, MixB).
  mix_atomic(Goal, Residue) :-
      make_atomic_residue(Goal, Residue).

/*%%  CLAUSES  %%%
*/

  mix_clauses(Head, [(Head:-Body)|RestCl], Residue) :-
      mix(Body, Residue).
  mix_clauses(Head, [_X|RestCl], MixBody) :-
      mix_clauses(Head, RestCl, MixBody).

/*%%  BAG OF RESIDUE  %%%
%%      Support for conditions when all the residues for a goal are
%%      required at the same time
%%      Example: side_effect predicates, after cut, meta-predicates
%%          and set predicates
*/

/*%%   Three such predicates are defined. They differ on whether they perform
%%   any postprocessing on the RESIDUES
*/

/*%    mix_bag(+GOALS, -RESIDUES)
%    mix_bag_eval(+GOALS, -RESIDUES)
%    mix_bag_eval_x(+GOALS, -RESIDUES)
%
%    mix_bag/2 returns RESIDUES without any postprocessing
%
%    mix_bag_eval/2 checks if the bag of residues is determinate (that
%        is the bag have one element). If so it evaluates the RESIDUES
%        if it is evaluable. The result of evaluation "true" or "fail"
%        is returned as the RESIDUES. Otherwise the bag is returned
%
%    mix_bag_eval_x/2 is the same as mix_bag_eval/3, except if the
%        evaluation of the residue results in a "fail" it fails.
*/

  mix_bag(Goals, Residue) :-
     variables_in(Goals, Vars),
     mix_bag_x(Goals, Vars, AllR),
     generate_unifications_for_bag(AllR, Vars, Residue).

     mix_bag_x(Goals, Vars, AllR) :-
	findall(sol(R,Vars), 
	    (nonvar(Vars),mix(Goals,R)),
	   AllR). 

  mix_bag_eval(Goal, Residue) :-
       mix_bag(Goal, MixGoal),
       eval_unifs_if_determinate(MixGoal, Residue).

  mix_bag_eval_x(Goal, Residue) :-
     mix_bag_eval(Goal, Residue),
     Residue /= fail.

/*% mix_bag(+HOWTOCUT, +GOAL, +BAG)
%    For the sake of compatibility only
*/

mix_bag(HowToCut, Goal, Bag) :-
    mix_bag(Goal, Bag).

/**  END OF MIX.PRO  ***/
