/*  Copyright (C) 1989,90  Arun Lakhotia    */
/* File: mix.pro
*  Author: Arun Lakhotia
*  Released:  Thu Jul 5 16:36:21 CDT 1990
*   $Log:	mix.pro,v $
**   Revision 1.3  90/07/05  17:22:45  arun
**   New Release
**   
**   Revision 1.2  89/04/17  17:23:22  arun
**   Added mechanism to detect loops
**   
**   Revision 1.1  89/04/16  01:25:30  arun
**   Initial revision
**         */

/* CHANGES
*      February 18, 1990 by AL
*       Fix for problem with evaluating in virtual database.
*       Replaced is_interpreted/1 by is_user_defined/1
*     February 23, 1990 by AL
*       Started working on FAULT.002. The problem with cut. At present
*       I am simply analyzing the potential changes. The potential
*       places for problem are marked **fault002.nn. The "nn"
*       represents the discussion number in file FAULT.002
*/ 

/*
*       The core of the partial evaluator houses here.
*       Created afresh after phasing out old code that served almost 2
*       years.
*       It takes care of ALL control structures and predicates of
*       Prolog. Code for meta predicates and set predicates is in the
*       files  set.pro and meta.pro
*/

/*
*    ENTRY POINTS
*      -   drive_mix(+GOAL, CLAUSES)
*      -   mix(+GOAL, -RESIDUE)
*
*    Other entry points for experimenting
*
*     The following returns one RESIDUE at a time. They differ on how
*     they handle cut and are needed in different contexts
*
*      -   mix_opaque(+GOAL, -RESIDUE)
*      -   mix_opaque(+GOAL, -RESIDUE, +ANCESTORS)
*      -   mix_transparent(+GOAL, -RESIDUE)
*      -   mix_transparent(+GOAL, -RESIDUE, +ANCESTORS)
*
*     The following returns a "bag" of all the RESIDUES. The parameter
*     HOWTOCUT tells how the cut is to be handled. It may be "opaque"
*     or "transparent"
*     
*      -   mix_bag(+HOWTOCUT, +GOAL, -RESIDUES)
*      -   mix_bag(+HOWTOCUT, +GOAL, -RESIDUES, +ANCESTORS)
*
*    Entry point USED BY THE DRIVER FOR SPECIALIZING INTERPRETERS. 
*    The driver performs one level of unfolding and passes the unfolded
*    CLAUSES.
*      -   mix_clauses(HEAD, CLAUSES, RESIDUE, +ANCESTORS)
*
*    In all the above ANCESTORS is a list of atoms that are copies of
*    goals in the ancestoral chain upto the given goal.
*    This list is generate internally. The form without ANCESTOR is
*    maintained to give an easy interface to when experimenting.
*/


%  drive_mix(+GOAL, -MIXEDCLAUSE)
%     A driver for mix for the specific purpose of specializing
%     interpreters. It performs one level of unfolding of the
%     interpreter GOAL. The other levels are suppressed by the
%     unfold rules
%     MIXEDCLAUSE is a clause from mixing GOAL. Only one clause is
%     returned at a time. Clauses due to hidden cut are returned too.

drive_mix(Goal, MixedClause) :-
    is_user_defined(Goal),
    all_clauses(Goal, Clauses),
    copy_hook(Goal, CopyGoal),
    mix_clauses(Goal, Clauses, MixedBody, [CopyGoal]), 
    filter_residue(MixedBody, ResBody, CutClauses-[]),
    member(MixedClause, [(Goal:-ResBody)|CutClauses]).


%  mix(+GOALS, -RESIDUE)
%       partially evaluates GOALS and returns the suspended computation
%       as RESIDUE
%       Used for partially evaluating queries (like when testing or
%       demonstrating ProMiX)

  mix(Goals, Residue) :-
       mix_opaque(Goals, Residue, []).

%  mix_opaque(+GOALS, -RESIDUE, +ANCESTORS)
%       partially evaluates GOALS and returns the suspended
%       computation as RESIDUE.
%       If GOALS contain 'cut' it is assumed to have local effect.
%       i.e. with in the scope of GOALS only.
%       If the goals before cut are evaluable it evaluates the cut ...
%       .... find the rest by playing with it. The RESIDUE does not
%       contain any indication about the encounter with cut
%   Is useful for handling the second cut in a body and cuts in
%      meta predicates where its effect is transparent
%   May be used handling top-level queries (mix/2)

  mix_opaque(Goal, Residue) :-
     mix_opaque(Goal, Residue, []).

  mix_opaque(Goal, Residue, Ancestors) :-
     mix_transparent(Goal, MixGoal, Ancestors),
     (  has_cut(MixGoal),
        !,
	mix_rest(MixGoal, Residue, Ancestors)
      ; Residue = MixGoal
     ).

     mix_rest(MixGoal, Residue, Ancestors) :-
	    evaluable_cut(MixGoal, BeforeCut, AfterCut),!,
	    mix_opaque(AfterCut, ResAfterCut, Ancestors),
	    merge_ands(BeforeCut, ResAfterCut, Residue).
     mix_rest(MixGoal, Residue, _Ancestors) :-
	    extract_residue(MixGoal, Residue).


%  mix_transparent(+GOALS, -RESIDUE, +ANCESTORS)
%       partially evaluates GOALS and returns the suspended
%       computation as RESIDUE.
%       If a cut is encountered in GOAL it is evaluated to cut
%       the choice points for conjunctions and an
%       indication to this effect is also returned in the RESIDUE.

  mix_transparent(Goals, Residue) :-
         mix_transparent(Goals, Residue, []).

  mix_transparent(Var, Residue, _Ancestors) :-
      var(Var),!,
      make_atomic_residue(call(Var), Residue).
  mix_transparent(true, Residue, _Ancestors):-
      !, make_atomic_residue(true, Residue).
  mix_transparent((X,Y), MixXY, Ancestors) :-!,
       mix_and(X, Y, MixXY, Ancestors).
  mix_transparent(X, Residue, Ancestors) :-
      is_control_predicate(X), !,
      mix_meta(X, Residue, Ancestors).
  mix_transparent(X, Residue, Ancestors) :-
      is_set_predicate(X), !,
      mix_set(X, Residue, Ancestors).
  mix_transparent(Goal, Residue, Ancestors) :-
      mix_atomic(Goal, Residue, Ancestors),
      ( has_cut(Residue),
        !
       ; true
      ).

    /* CONJUNCTIVE GOALS  */

  mix_and(X, Y, MixXY, Ancestors) :-
      ( var(X)
      ; is_side_effect(X)
      ; is_extra_logical(X)
      ),!,
       mix_bag(transparent, X, MixX, Ancestors),
       mix_bag(transparent, Y, MixY, Ancestors), 
       merge_ands(MixX, MixY, MixXY).

      % no need to cut choice points since X cannot have a !
  mix_and(GoalA, GoalB, MixAB, Ancestors) :-
      mix_transparent(GoalA, MixA, Ancestors),
      mix_conj(MixA, GoalB, MixAB, CutAnd, Ancestors),
      % If one of the conjuncts has a cut, cut conjuncts 
    ( CutAnd = cut_and,
        !
      ; true
      ).

     mix_conj(MixX, Y, MixXY, cut_and, _Ancestors) :-
	  has_cut(MixX),!,
	  merge_ands(MixX, Y, MixXY).   % **fault002.1

     %** fault002.3
     %  The following is a *hack* to inhibit backtracking over residue
     %  containing hidden cut. Backtracking generates multiple
     %  identical "extension" clauses.

     mix_conj(MixX, Y, MixXY, cut_and, Ancestors) :-
	  has_hidden_cut(MixX),!,
	  mix_bag(transparent, Y, MixY, Ancestors),
	  merge_ands(MixX, MixY, MixXY),
	  ( has_cut(MixXY) 
            -> CutAnd = cut_and
            ;  CutAnd = no
          ).
     mix_conj(MixX, Y, MixXY, CutAnd, Ancestors) :-
	  mix_transparent(Y, MixY, Ancestors),
	  merge_ands(MixX, MixY, MixXY),
	  ( has_cut(MixXY) 
            -> CutAnd = cut_and
            ;  CutAnd = no
          ).

    /*   ATOMIC GOALS   */

  mix_atomic(!, Residue, _Ancestors) :-!, 
      make_cut_residue(!, Residue).
  mix_atomic(X, Residue, _Ancestors) :-
      immediately_evaluable(X), !, 
      evaluate(X),
      make_atomic_residue(true, Residue).
  mix_atomic(Head, MixB, Ancestors) :-
      computation_is_looping(Head, Ancestors),!,
      MixB = Head.
  mix_atomic(Head, MixB, Ancestors) :-
      is_user_defined(Head),
      should_unfold(Head), !,
      copy_hook(Head, Parent),
      all_clauses(Head, Clauses),
      mix_clauses(Head, Clauses, MixB, [Parent|Ancestors]).
  mix_atomic(Goal, Residue, _Ancestors) :-
      make_atomic_residue(Goal, Residue).

     /*  CLAUSES  */

  mix_clauses(Goal, [(Head:-Body)|RestCl], Residue, Ancestors) :-
      copy_hook(Goal, Head),
      mix_transparent(Body, MixBody, Ancestors),
      ( has_cut(MixBody),
        !,          /*  stop clause selection */
	commit_clause(Goal, Head, MixBody, RestCl, Residue, Ancestors)
      ; Goal = Head,
        Residue = MixBody
      ).
  mix_clauses(Head, [_X|RestCl], MixBody, Ancestors) :-
      mix_clauses(Head, RestCl, MixBody, Ancestors).

     %%   Processing of clause after encountering a cut done here
     commit_clause(Goal, Goal, MixBody, _RestCl, Residue, Ancestors) :-
	    % if residue before cut is evaluable, commit this clause
	     evaluable_cut(MixBody, BeforeCut, AfterCut),!,
	     mix_opaque(AfterCut, ResAfterCut, Ancestors),
	     merge_conjunctions(BeforeCut, ResAfterCut, Residue).
     commit_clause(Goal, Head, WMixBody, RestCl, Residue, _Ancestors) :-
	    % else stop partial evaluation and return 
	    % the partially evaluated clause and the set of clauses after it
	    %  *** This part would be improved in later versions ***
	     extract_residue(WMixBody, Body),
             make_clauses_with_cut(Goal, [(Head:-Body)|RestCl], Residue).

  %%  BAG OF RESIDUE  
  %%      Support for conditions when all the residues for a goal are
  %%      required at the same time
  %%      Example: side_effect predicates, after cut, meta-predicates
  %%          and set predicates

  %%   Three such predicates are defined. They differ on whether they perform
  %%   any postprocessing on the RESIDUES
  %
  %    mix_bag(+HOWTOCUT, +GOALS, -RESIDUES, +ANCESTORS)
  %    mix_bag_eval(+HOWTOCUT, +GOALS, -RESIDUES, +ANCESTORS)
  %    mix_bag_eval_x(+HOWTOCUT, +GOALS, -RESIDUES, +ANCESTORS)
  %
  %    HOWTOCUT can be "opaque" or "transparent". See description above
  %      about the two
  %
  %    mix_bag/4 returns RESIDUES without any postprocessing
  %
  %    mix_bag_eval/4 checks if the bag of residues is determinate (that
  %        is the bag have one element). If so it evaluates the RESIDUES
  %        if it is evaluable. The result of evaluation "true" or "fail"
  %        is returned as the RESIDUES. Otherwise the bag is returned
  %
  %    mix_bag_eval_x/4 is the same as mix_bag_eval/4, except if the
  %        evaluation of the residue results in a "fail" it fails.


  mix_bag(HowToCut, Goals, Residue) :-
     mix_bag(HowToCut, Goals, Residue, []).

  mix_bag(HowToCut, Goals, Residue, Ancestors) :-
     variables_in(Goals, Vars),
     mix_bag_x(HowToCut, Goals, Vars, AllR, Ancestors),
     generate_unifications_for_bag(AllR, Vars, Residue).

     mix_bag_x(opaque, Goals, Vars, AllR, Ancestors) :-
	findall(sol(R,Vars), 
	    (nonvar(Vars),mix_opaque(Goals,R,Ancestors)),
	   AllR). 

     mix_bag_x(transparent, Goals, Vars, AllR, Ancestors) :-
	findall(sol(R,Vars), 
	    (nonvar(Vars),mix_transparent(Goals,R,Ancestors)),
	   AllR).

  mix_bag_eval(HowToCut, Goal, Residue, Ancestors) :-
       mix_bag(HowToCut, Goal, MixGoal, Ancestors),
       eval_unifs_if_determinate(MixGoal, Residue).

  mix_bag_eval_x(HowToCut, Goal, Residue, Ancestors) :-
     mix_bag_eval(HowToCut, Goal, Residue, Ancestors),
     Residue /= fail.

/**  END OF MIX.PRO  ***/
