/*  Copyright (C) 1989,90  Arun Lakhotia    */
/* File: residue.pro
*  Author: Arun Lakhotia
*  Released:  Thu Jul 5 16:36:25 CDT 1990
*   $Log:	residue.pro,v $
**   Revision 1.2  90/07/05  17:22:56  arun
**   New Release
**   
**   Revision 1.1  89/04/16  01:25:31  arun
**   Initial revision
**         */

/* CHANGES
*    February 23, 1990: Initiated changes for FAULT.002 related to
*    cut. Am analyzing the problem at present. The clauses that
*    potentially contribute to it or may be eventually affected are
*    marked as **fault002.nn where "nn" is the point number in
*    FAULT.002 file in the discussion.
*/

/*
*   The mix code is transparent to the structure of RESIDUE. This is
*   acheived by abstracting the structure, and performing all the
*   tests, creation, and manipulation of residue through routines
*   provided here.
*
*   The benefits of this approach are obvious and need not be
*   elaborated upon.
*
**  DATA ABSTRACTION FOR RESIDUE AND NOTES ON HANDLING CUT
*
*   A RESIDUE has one of the following structures
*
*        '$$with_cut'(BEFORECUT, GOALS, AFTERCUT)
*        '$$with_cut'(GOALS)
*        '$$hidden_cut'(GOAL, CLAUSES)
*          GOALS      : Prolog Goals
*   The '$$with_cut'/[1,3] forms are used for residue of goals in the
*   same clause body. They encapsulate GOALS that contain a '!' and
*   have not been evaluable.
*
*   '$$with_cut'(BEFORECUT, GOALS, AFTERCUT) form is used for tracking
*   cuts in a conjunctive sequence. BEFORECUT is the "residue" for goals
*   before the GOALS containing cut. AFTERCUT are the goals AFTER the
*   GOALS containing '!'.
*
*   The '$$with_cut'/1,3 structures are used to pass the information
*   that a cut was found to the SCOPE of the '!'
*
*   CUTTING GOALS
*
*       When the residue of a SUBGOAL 'has_cut'/1 a '!' is executed to
*   inhibit backtracking over previous goal. and a residue $$with_cut
*   is returned to the outer levels. (See mix_and/2, mix_meta_if/2 etc)
*
*   CUTTING CLAUSES
*
*   If the RESIDUE of a clause body has_cut/1
*        if it is of the form '$$with_cut'/3
*           and BEFORECUT is an evaluable_residue/1 during partial evaluation
*           and GOALS = "!", the clause selection is terminated
*               the presence of cut is forgotten
*        else
*         % we have in hand GOALS that has_cut/1 that cannot be evaluated.
*         the "remains" of this clause and REST_CLAUSES are packed in a
*         '$$hidden_cut'/2 structure and returned as RESIDUE
*
*  RESERVED STRUCTURES
*
*   The programs being partially evaluated should not contain the following
*   structures.
*      $$hidden_cut/2
*      $$with_cut/1,3
*      $$unifs/1    -- used to remember unifications generated internally
*/

%  make_atomic_residue(+GOAL, -RESIDUE)
%      create RESIDUE for atomic GOAL

  make_atomic_residue(Goal, Goal).

%   make_meta_residue(+GOAL, -RESIDUE)
%      create RESIDUE for a meta GOAL

  make_meta_residue(MetaRes, MetaRes).

%   make_clauses_with_cut(+GOAL, +CLAUSES, -HIDDEN_CUT)
%       creates residue with HIDDEN_CUT when unfolding of GOAL unfolds
%       a cut.

  make_clauses_with_cut(Goal, Clauses, '$$hidden_cut'(NewGoal,NewClauses)) :-
      gensym(NewNum),
      Goal =.. [Functor|Args],
      append_atoms(Functor, NewNum, NewFunctor),
      NewGoal =.. [NewFunctor|Args],
      rename_clauses_heads(Clauses, NewFunctor, NewClauses).

      rename_clauses_heads([], _NewFunctor, []).
      rename_clauses_heads([(H:-B)|Cls], NewFunctor, [(NewH:-B)|NewCls]) :-
          H =.. [_|Args],
          NewH =.. [NewFunctor|Args],
          rename_clauses_heads(Cls, NewFunctor, NewCls).

      :- dynamic '$$gensym'/1.

      gensym(Sym) :- 
	   (retract('$$gensym'(X)); X = 0),!,
	   Num is X+1,
	   asserta('$$gensym'(Num)),
	   append_atoms('_extension_', Num, Sym).

% make_cut_residue(+GOALS, -WITH_CUT)
%   creates a WITH_CUT structure with GOALS as residue

  make_cut_residue(!, '$$with_cut'(true,!,true)).
  make_cut_residue((RX->RY;RZ), '$$with_cut'((RsX->RsY;RsZ))) :-!,
     extract_residue(RX, RsX),
     extract_residue(RY, RsY),
     extract_residue(RZ, RsZ).
  make_cut_residue((RX;RY), '$$with_cut'((RsX;RsY))) :-!,
     extract_residue(RX, RsX),
     extract_residue(RY, RsY).

%  merge_ands(+RESIDUE1, +RESIDUE2, -RESIDUE)
%    creates RESIDUE12 representing a conjunction of RESIDUE1 and RESIDUE2

  merge_ands(fail, _X, fail) :-!.
  merge_ands(MixX, MixY, _MixXY) :-             %** fault002.2
       has_cut(MixX), 
       has_cut(MixY),
       fatal_error('merge_ands/3 received both residues with cut').
  merge_ands('$$with_cut'(Before,Cut,After), MixY, 
                         '$$with_cut'(Before,Cut,AfterCut)):-
       !,
       extract_residue(MixY, ResY),
       merge_conjunctions(After, ResY, AfterCut).
  merge_ands(MixX, '$$with_cut'(Before,Cut,After),
                          '$$with_cut'(BeforeCut,Cut,After)):-
       !,
       extract_residue(MixX, ResX),
       merge_conjunctions(ResX, Before, BeforeCut).
  merge_ands(MixX, MixY, '$$with_cut'(MixXandY)) :-
       (  has_cut(MixX)
        ; has_cut(MixY)
       ),!,
       extract_residue(MixX, ResX),
       extract_residue(MixY, ResY),
       flatten_conjunction((ResX,ResY), MixXandY).
  merge_ands(MixX, MixY, MixXandY) :-
       flatten_conjunction((MixX,MixY), MixXandY).

   /*   SUPPORT PROCEDURES FOR HANDLING RESIDUE */

%  has_hidden_cut(+RESIDUE)
%    succeeds if RESIDUE had a hidden cut

  has_hidden_cut('$$hidden_cut'(_X,_Xs)).

%  has_cut(+RESIDUE)
%    succeeds if RESIDUE has cut

  has_cut('$$with_cut'(_)).
  has_cut('$$with_cut'(_,_,_)).

%  extract_residue(+RES_STRUCTURE, -GOALS)
%    extracts the GOALS in the RES_STRUCTURE.

  extract_residue('$$with_cut'(Goals), Goals) :-!.
  extract_residue('$$with_cut'(Before,Cut,After), Goals) :-!,
        flatten_conjunction((Before,Cut,After), Goals).
  extract_residue(Goals, Goals).

%  evaluable_cut(+RESIDUE, -BEFORECUT, -AFTERCUT)
%   suceeds if conditions for evaluating a cut in the residue are
%   satisfied. Returns the residue BEFORECUT and the goals AFTERCUT

  evaluable_cut(!, true, true).
  evaluable_cut('$$with_cut'(BeforeCut,!,AfterCut), BeforeCut, AfterCut) :-
      evaluable_residue(BeforeCut).

 /*    SUPPORT TO PROCESS BAG OF RESIDUES  */

% generate_unifications_for_bag(+SOLUTIONS, +VARS, -BAGS)

   generate_unifications_for_bag([], _Vars, fail):-!.
   generate_unifications_for_bag([Sol], Vars, Res) :-!,
	 gx(Sol, Vars, Res).
   generate_unifications_for_bag([Sol|Us], Vars, Bag) :-
	 gx(Sol, Vars, Res),
	 generate_unifications_for_bag(Us, Vars, Ts),
         merge_ors(Res, Ts, Bag).

  % gx(+SOLUTION, +VARS, -RESIDUE)


   gx(sol(Res,U), Vars, Residue) :-!,
       generate_unifications(Vars, U, Unifications),
       translate_to_unifs(Unifications, Unifs),
       merge_ands(Unifs, Res, Residue).
   gx(Sol,_,_) :-
      % A trap for errors during development */
       writeln(['Error: Incorrect structure ', Sol,' to filter_residue_bag ']).


% generate_unifications(+VARS, +VALUES, -UNIFICATIONS)
%    if the VALUE corresponding to a VAR is bound it generates a
%    unification for it. If it a variable the two variables are
%    unified.

   generate_unifications([], [], true).
   generate_unifications([Var|XVars], [CopyVar|CVars], TUnifs) :-
	var(CopyVar), !,
	Var = CopyVar,
	generate_unifications(XVars, CVars, TUnifs).
   generate_unifications([Var|XVars], [CopyVar|CVars], Unifs) :-
	generate_unifications(XVars, CVars, TUnifs),
	merge_conjunctions((Var = CopyVar), TUnifs, Unifs).

   % translate_to_unifs(+UNIFS, -UNIF_STRUCTURE)

   translate_to_unifs(true, true):-!.
   translate_to_unifs(Unifications, '$$unifs'(Unifications)).

  '$$unifs'(X) :- X.   /*  Just in case you wish to execute $$unif/1 */

% evaluable_residue(+BAG)
%    Tests when a BAG has residue that is evaluable. The conditions
%    at present may be too strong. They may be relaxed after
%    gaining proper experience.


  evaluable_residue('$$unifs'(_X)).
  evaluable_residue(true).
  evaluable_residue(fail).

% eval_unifs_if_determinate(+BAG, -RESIDUE)
%   evaluates a BAG if determinate. The BAG itself is RESIDUE
%   otherwise. 


  eval_unifs_if_determinate('$$unifs'(Unifs), Residue) :-!,
       call_hook(Unifs),
       make_atomic_residue(true, Residue).
  eval_unifs_if_determinate(Bag, Bag).


% evaluable_bag(+BAG)
%    succeeds if all the residues in BAG are evaluable

  evaluable_bag((R;Rs)) :- !,
	 evaluable_residue(R),
	 evaluable_bag(Rs).
  evaluable_bag(R) :- evaluable_residue(R).

% eval_first_residue(+BAG, -RESIDUE)
%   Hard to comment..
%    This is  used in handling not/1.
%    May also be used for handling  condition part of
%    if-then-else. (It isn't at present)
%
%   NOTE: The following codes "knows" the structure of residue. If this
%     structure is changed, the code should be changed accordingly

  eval_first_residue(fail, fail):-!.
  eval_first_residue(true, true):-!.
  eval_first_residue((true;_Y), Residue) :-!,
        make_atomic_residue(true, Residue).
  eval_first_residue(Bag, Bag).

/*  SUPPORT TO EXTRACT HIDDEN CLAUSES CONTAINING CUT   */

%  filter_residue(+RESIDUE, -FILTERED_RESIDUE, -CLAUSES)
%     extracts a difference list of CLAUSES left hidden in RESIDUE and
%     return FILTERED_RESIDUE 

  filter_residue(Residue, FResidue, Clauses-Tail) :-
      filter_residue(Residue, FResidue, Clauses, Tail).

  filter_residue(X, X, Tail, Tail) :- var(X),!.
  filter_residue('$$hidden_cut'(Goal,Clauses), Goal, Head, Tail) :-!,
    filter_hidden_residues(Clauses, Head, Tail).
  filter_residue('$$unifs'(X), X, Tail, Tail):-!.
  filter_residue((GoalB,GoalC), FGoalBC, Head, Tail) :-!, 
     filter_residue(GoalB, FGoalB, Head, TailX), 
     filter_residue(GoalC, FGoalC, TailX, Tail),
     merge_conjunctions(FGoalB, FGoalC, FGoalBC).
  filter_residue(MetaPred, FMetaPred, Head, Tail) :-
    meta_property(_,MetaPred, MetaProperty),!,
    functor(MetaPred, Functor, Arity),
    functor(FMetaPred, Functor, Arity),
    filter_m(MetaPred, MetaProperty, FMetaPred, 1, Arity, Head, Tail).
  filter_residue(Atom, Atom, Tail, Tail).

/*  filter_residue of the meta-arguments of meta-predicates */

  filter_m(_MetaPred, _MetaProperty, _FMetaPred, ArgNum, Arity, Tail, Tail) :-
      ArgNum > Arity,!.
  filter_m(MetaPred, MetaProperty, FMetaPred, ArgNum, Arity, Head, Tail) :-
      arg(ArgNum, MetaPred, ArgN),
      arg(ArgNum, MetaProperty, PropertyArg),
      arg(ArgNum, FMetaPred, FMetaArg),
      (is_meta_argument(PropertyArg)  
	-> filter_residue(ArgN, FMetaArg, Head, Tailx)
	; FMetaArg = ArgN, Head=Tailx),
      ArgNum1 is ArgNum+1,
      filter_m(MetaPred, MetaProperty, FMetaPred, ArgNum1, Arity, Tailx, Tail).


   filter_hidden_residues([], Head, Head).
   filter_hidden_residues([C|Cs], [FC|Head], Tail) :-
	filter_residue(C, FC, Head, Mid),
	filter_hidden_residues(Cs, Mid, Tail).

/*   END OF RESIDUE.PRO  */
