/*%   File   : dynaload.pl
%   Author : Richard A. O'Keefe
%   Purpose: load a file of clauses as dynamic
%   Started:  9:41 pm, 27 August 1986.
%   Done by: 10:19 pm same day (includes testing, 3 minutes other work)
%   Version: Quintus Prolog release 2.0Beta.
*/

/*  This file defines one operation:

	dynaload(Files)

    Files is either a file name or a list of file names.  A .pl suffix
    is **NOT** added; only compile/1 and consult/1 and so on do that.
    Each file should contain ordinary Prolog clauses, and nothing else.
    Grammar rules are allowed; we call expand_term/2, and this means
    that transformations of your own can be done by term_expansion/2.

    When you dynaload() a file, all the clauses which came from that
    file before are erased, and NO OTHERS.  This is pretty much like
    consult/1 used to be in Dec-10 Prolog and C Prolog; all the
    predicates affected by dynaload/1 are :- dynamic and :- multifile
    by default.  There is an important difference: dynaload()ing a
    file twice gives you a single copy of its contents.  Dynaload/1
    thus replaces consult/1 and reconsult/1, except that it permits
    neither :-commands nor ?-queries in the files it reads.

    This file was written in a race against the clock, and to test the
    idea.  So this prototype of it does not support the Quintus Prolog
    module system.  Instead, every clause is loaded into the user:
    module, which is where consult/1 would have put it prior to 2.0.
*/

:- module(dynaload, [
	dynaload/1
   ]).
:- use_module(library(files), [
	open_file/3
   ]).

:- dynamic
	file_originated_dbref/2.

	
dynaload([File|Files]) :- !,
	dynaload_list([File|Files]).
dynaload(File) :-
	(   atom(File) ->
	    open_file(File, read, Stream),
	    purge_clauses_from(File),
	    repeat,
		read(Stream, RawTerm),
		expand_term(RawTerm, Clause),
		store_clause(Clause, File),
		Clause = end_of_file,
	    !,
	    close(Stream)
	;   format(user_error, '~N! dynaload(~w) : argument is not an atom~n',
		[File]),
	    fail
	).


dynaload_list([]).
dynaload_list([File|Files]) :-
	dynaload(File),
	dynaload_list(Files).


purge_clauses_from(File) :-
	retract(file_originated_dbref(File, DBref)),
	erase(DBref),
	fail
    ;	true.


store_clause(end_of_file, _).
store_clause(:-(Command), File) :- !,
	current_output(Stream),
	set_output(user_error),
	format('~N! dynaload/1 found a command in file ~w~n', [File]),
	user:portray_clause(:-(Command)),
	set_output(Stream).
store_clause(?-(Query), File) :- !,
	current_output(Stream),
	set_output(user_error),
	format('~N! dynaload/1 found a query in file ~w~n', [File]),
	user:portray_clause(?-(Query)),
	set_output(Stream).
store_clause(Clause, File) :-
	user:assert(Clause, DBref),
	assert(file_originated_dbref(File,DBref)).


