/*  Copyright (C) 1989,90  Arun Lakhotia    */
/* File: prolog_metai.pro
*  Author: Arun Lakhotia
*  Released:  Thu Jul 5 16:46:01 CDT 1990
*   $Log:	prolog_metai.pro,v $
**   Revision 1.2  90/07/05  17:23:42  arun
**   New Release
**   
**   Revision 1.1  89/04/16  01:35:17  arun
**   Initial revision
**         */

/*
*   CHANGES
*    FEBRUARY 18, 1990  by AL
*      Incorporated changes to fix the fault when using virtual database.
*      The primitive is_primitive/1 and is_user_defined/1 are replaced
*      for the previous is_interpreted/1, is_compiled/1 etc.
*/
/* 
*   A PROLOG META-INTERPRETER FOR "FULL" PROLOG
*
*   The interpreter implements cut, meta-logical predicates (such as
*   not/1, call/1, ;/2) and the set predicates such as (setof/3,
*   bagof/3, findall/3)
*
*   Though it uses the equivalent predicates of the native interpreter
*   to implement these features, it may be noticed that all
*   user-defined predicates are interpreted by this interpreter. The
*   interpreter is therefore of interest
*   
*      1) for using as a skeleton for developing meta-programs such as
*         a tracer and partial evaluator.
*      2) when the "data" Prolog program being interpreted is not
*         stored in the Prolog database but instead is saved as say
*         $$program(X,Y)" facts. 
*         Such a situation arises when an implementation inhibits or
*         restricts the use of clause/2 predicate
*
*   To aid in the second point the interpreter uses unfold_clause/2
*   (instead of clause/2) to get the definition of a clause from the
*   database. Appropriate definition of this clause should be provided
*   by the user.
*
*   Predicates used by this interpreter and whose definitions may
*   depend on the underlying Prolog implementations are:
*
*       is_primitive/1, is_user_defined/1, evaluate/1
*
*   The behavior of these predicates may be obvious from their names
*   and their usage. They may be interfaced with appropriate system
*   predicates.
*
*   Other predicates used but not defined here are:
*
*       is_set_predicate/1 and is_meta_predicate/1
*
*   This predicates test if their arguments belong to the specific
*   class of predicates. These predicates are easy to define; and are
*   also defined elsewhere in the thesis.
*/

%  prolog_goal(+GOALS).
%        evaluate Prolog GOALS

  prolog_goal(Goals) :-
     solve_goal(Goals, Cut, AfterCut),
     (Cut == !, !, solve_goal(AfterCut)
	       ; true).

%  solve_goal(+GOALS, -CUT, -AFTERCUT)
%    evaluates a sequence of GOALS till it either finds a CUT
%    or the sequence is exhausted.
%    In the former case it binds CUT to `!' otherwise to `true'
%    AFTERCUT are the conjunctive goals that follow the cut
%    If `!' is the last atom in GOALS then AFTERCUT is bound to `true'

  solve_goal(!, !, true) :-!.
  solve_goal((!, AfterCut), !, AfterCut):-!.
  solve_goal((GoalA,GoalB), Cut, AfterCut) :-!,
      solve_goal(GoalA, CutA, AfterCutA),
      solve_conjunct(CutA, AfterCutA, GoalB, Cut, AfterCut).
  solve_goal(Goal, Cut, AfterCut) :-
      is_meta_predicate(Goal),!,
      solve_meta_goal(Goal, Cut, AfterCut).
  solve_goal(Goal, true, true) :-
      is_set_predicate(Goal),!,
      solve_set_goal(Goal).
  solve_goal(Goal, no, true) :-
      solve_atomic_goal(Goal).

%  solve_conjunct(+PREV_CUT, +PREV_AFTERCUT, +GOALS, -CUT, -AFTERCUT)
%     processess `!' in a conjunction
%     PREV_CUT is `!' if the conjunctive goals before GOALS has cut
%     PREV_AFTERCUT is the sequence of goals after this cut but
%         before GOALS
%     If the previous goals had a cut then GOALS is not evaluated
%         and is returned as AFTERCUT by conjuncting with PREV_AFTERCUT
%     else it is evaluated

  solve_conjunct(!, AfterCut, Goals, !, (AfterCut,Goals)) :-!.
  solve_conjunct(_, _,      Goals, Cut, AfterCut) :-
      solve_goal(Goals, Cut, AfterCut).

%  solve_atomic_goal(+GOAL)
%    executes an atomic GOAL
%    Only goals that can be interpreted are explicitly interpreted
%    Built_in primitives or compiled goals cannot be interpreted
%    and are hence evaluated directly by the system
%    The execution is transparent to any cut encountered in the evaluation
%    of GOAL

  solve_atomic_goal(Head) :-
      is_user_defined(Head),!,  % changed 2/18/90 AL
      unfold_clause(Head, Body),
      solve_goal(Body, Cut, AfterCut),
      (Cut = !,
	 !,
	 prolog_goal(AfterCut)
      ; true).
  solve_atomic_goal(Goal) :-
      is_primitive(Goal),      % changed 2/18/90 AL
      evaluate(Goal).


% solve_meta_goal(+META_PREDICATE, -CUT, +AFTERCUT)
%     Interprets a META_PREDICATE

  solve_meta_goal(call(X), Cut, AfterCut) :- solve_goal(X, Cut, AfterCut).
  solve_meta_goal((A;B), Cut, AfterCut) :- 
	 solve_goal(A, Cut, AfterCut)
       ; solve_goal(B, Cut, AfterCut).
  solve_meta_goal((A->B;C), Cut, AfterCut) :- 
       solve_goal(A, CutA, AfterCutA) 
       -> (CutA = !,
	   !,
	   Cut = !, 
	   AfterCut = (AfterCutA ->B;C)
	  ; solve_goal(B, Cut, AfterCut))
       ; solve_goal(C, Cut, AfterCut).
  solve_meta_goal(\+X, no, true) :- \+ prolog_goal(X).

%  solve_set_goal(+SET_PREDICATE)
%      Interprets a SET_PREDICATE

  solve_set_goal(findall(X,Y,Z)) :- findall(X, prolog_goal(Y), Z).
  solve_set_goal(bagof(X,Y,Z)) :- bagof(X, prolog_goal(Y), Z).
  solve_set_goal(setof(X,Y,Z)) :- setof(X, prolog_goal(Y), Z).


