/*  Copyright (C) 1989,90  Arun Lakhotia    */
/* File: clause.pro
*  Author: Arun Lakhotia
*  Released:  Thu Jul 5 16:37:43 CDT 1990
*   $Log:	clause.pro,v $
**   Revision 1.2  90/07/05  17:23:33  arun
**   New Release
**   
**   Revision 1.1  89/04/16  01:23:05  arun
**   Initial revision
**         */
/* CHANGES
*     February 18, 1990 by AL
*     To fix problem with evaluating goal in the virtual database
*       Changed the definition of unfold_clause/2 by removing the
*       first clause that was trapping sys/1 goals. Hope this will not
*       break anything
*/

/*
*   Purpose: Collect utilities related to operations
*      on Prolog clause structure
*
*/

/*%  and_to_list(?ANDS, ?LIST)
%      Creates a LIST from a sequence of ANDS.
%      Order of elements and their numbers are maintained.
%      Works both ways
*/

  and_to_list(true, []) :-!.
  and_to_list(X,[X]) :- \+X=(_,_),!.
  and_to_list((X,Y),[X|Xs]) :- and_to_list(Y,Xs).

/*%  or_to_list(?ORS, ?LIST)
%      Creates a LIST from a sequence of ORS.
%      Order of elements and their numbers are maintained.
%      Works both ways
*/

  or_to_list(fail, []) :-!.
  or_to_list(X, [X]) :-!.
  or_to_list((X;Ors), [X|Z]) :-
       or_to_list(Ors, Z).

/*% flatten_conjunction(+SEQUENCE, -FLATSEQ)
%   Flattens nested conjunctions and removes redundant `true's
%   from SEQUENCE to give FLATSEQ
*/

  flatten_conjunction(X, X) :-
      var(X),!.
  flatten_conjunction((H:-B), (H:-NewB)) :-
      flatten_conjunction(B, NewB), !.
  flatten_conjunction(A, Af) :- 
      flatten_conjunction(A, true, Af).

    flatten_conjunction(X, Pf, Af) :-
	var(X),!,
	merge_conjunctions(X, Pf, Af).
    flatten_conjunction((A,B), Pf, Af) :- !,
	flatten_conjunction(B, Pf, Bf),
	flatten_conjunction(A, Bf, Af).
    flatten_conjunction(Pred, Pf,  Af) :-
	 meta_property(_, Pred, Property),!,
	 functor(Pred, Functor, Arity),
	 functor(FPred, Functor, Arity),
	 flatten_meta_arguments(Arity, Pred, Property, FPred),
	 merge_conjunctions(FPred, Pf, Af).
    flatten_conjunction(Pred, Pf, Af) :-
	 merge_conjunctions(Pred, Pf, Af).

    flatten_meta_arguments(0, _Pred, _Property, _FPred) :-!.
    flatten_meta_arguments(NthArg, Pred, Property, FPred) :-
	  arg(NthArg, Pred, A),
	  arg(NthArg, Property, P),
	  arg(NthArg, FPred, F),
	  N1 is NthArg -1,
	 (is_meta_argument(P)
	    -> flatten_conjunction(A, true, F)
	    ; F = A),
	 flatten_meta_arguments(N1, Pred, Property, FPred).

/*%  merge_conjunctions(+X, +Y, -XANDY)
%      If X and Y are goals XANDY is their conjunction.
%      XANDY is in canonical form such that it does not
%      have redundant 'true'
*/

  merge_conjunctions(A, B, A) :-
       nonvar(B),
       B = true,!.
  merge_conjunctions(A, B, B) :- 
       nonvar(A),
       A = true,!.
  merge_conjunctions(A, B, (A,B)).

/*% all_clauses(+GOAL, -CLAUSE_BODY_PAIRS)
%      where CLAUSE_BODY_PAIRS is a list of (HEAD, BODY)
%            where HEAD' :- BODY' is a clause in the Database
%            and (HEAD,BODY) is the resolvent from unifying
%            HEAD' with GOAL.
*/

  all_clauses(Goal, ClauseBodyPairs) :-
	bagof((Goal :- Body), unfold_clause(Goal, Body), ClauseBodyPairs).

/*%  unfold_clause(+HEAD, -BODY)
%     the same as clause/2, except it checks if the HEAD
%     corresponds to a system defined predicate.
%     In some systems clause/2 on system defined preicates gives
%     an error, rather than failing gracefully. An advance check
%     avoids this error message.
%   The procedure unfold_data/2 provides a handle for portability.
%   For systems that are completely compiled, the data may be
%   stored in a different form.
*/

   unfold_clause(A, B) :- unfold_data(A, B).

/*%  clause_head(+CLAUSE, -HEAD)
%     HEAD is the head atom of CLAUSE
*/

  clause_head((Head:-_Body), Head) :-!.
  clause_head(Head, Head).

/*%  clause_body(+CLAUSE, -HEAD)
%     BODY is the body of CLAUSE
*/

  clause_body((_Head :- Body), Body):-!.
  clause_body(_Head, true).

/*%  split_clause(+CLAUSE, ?HEAD, ?BODY)
%     splits a CLAUSE into its HEAD and BODY
%     If it is a fact (ie a term) the BODY is `true'
*/

  split_clause((Head:-Body), Head, Body) :-!.
  split_clause(Fact, Fact, true).

/*%   term_to_clause(+TERM, +CLAUSE)
%       If TERM is a clause then CLAUSE.
%       else CLAUSE is (TERM:-true)
*/

  term_to_clause((A:-B), (A:-B)) :- !.
  term_to_clause(A,(A:-true)).

/*%%
%%  END OF CLAUSE
%%*/
