/*  Copyright (C) 1989,90  Arun Lakhotia    */
/* File: util.pro
*  Author: Arun Lakhotia
*  Released:  Thu Jul 5 16:38:10 CDT 1990
*   $Log:	util.pro,v $
**   Revision 1.3  90/07/05  17:23:13  arun
**   New Release
**   
**   Revision 1.2  89/04/17  17:25:35  arun
**   Added predicates to compare two terms.
**   like most_general/2, instance_of/2, ..
**   
**   Revision 1.1  89/04/16  01:23:10  arun
**   Initial revision
**         */
/*
*  Purpose: general purpose utility predicates
*/

/*%   +TERM1 /= +TERM2
%        TERM1 does not unify with TERM2
*/
   X /= Y :- \+ X = Y.

/* instance_of(+TERM1, +TERM2)
*      TERM1 is an instance of TERM2
*/

   instance_of(Term1,Term2) :-
       copy_hook(Term1,Term3),
       symbolvars(Term3,0,_),
       unifiable(Term2,Term3).

/* more_general(+TERM1, +TERM2)
*     TERM1 is more general than TERM2
*/
   more_general(Term1,Term2) :- instance_of(Term2,Term1).

/* less_general(+TERM1, +TERM2)
*      TERM1 is less general than TERM2
*/

   less_general(Term1,Term2) :- instance_of(Term1,Term2).

/* unifiable(+TERM1, +TERM2)
*      TERM1 is unifiable with TERM2
*/
   unifiable(Term1, Term2) :- \+ \+ Term1=Term2.

/* equal(+TERM1, +TERM2)
*      TERM1 is structurally equal to TERM2
*/
   equal(Term1, Term2)     :- Term1 == Term2.

/* most_general(+TERM1, +TERM2, -TERM)
*      TERM is the most general of the two terms TERM1 and TERM2
*/
   most_general(Term1,Term2,Term1) :- instance_of(Term2,Term1).
   most_general(Term1,Term2,Term2) :- instance_of(Term1,Term2).

/*%   extra_variables(+TERM1, +TERM2, -EXTRA_VARIABLES)
%        EXTRA_VARIABLES used in TERM1 but not in TERM2
*/

   extra_variables(Term1, Term2, Difference) :-
	  variables_in(Term1, Vars1),
	  variables_in(Term2, Vars2),
	  difference_s(Vars1, Vars2, Difference).

/*%  variables_in(+TERM, -SET)
%       SET of variables in TERM. (The set is represented as a list)
*/

   variables_in(Term, Set) :-
	  variables_in(Term, [], Set).

   variables_in(V, OldSet, NewSet) :- 
       var(V),!,
       (member_s(V, OldSet)
	   -> NewSet = OldSet
	    ; NewSet = [V|OldSet]).
   variables_in(Term, InSet, OutSet) :- 
       Term =.. [_Functor|Args],
       variables_in_x(Args, InSet, OutSet).

      variables_in_x([], InSet, InSet).
      variables_in_x([X|Xs], InSet, OutSet) :- 
	  variables_in(X, InSet, OutSet1), 
	  variables_in_x(Xs, OutSet1, OutSet).

/*% difference_s(+LIST1, +LIST2, -DIFFERENCE)
%      DIFFERENCE of "structures" in LIST1 and LIST2.
%      Structures in LIST1 but not in LIST2.
*/

   difference_s([], _List2, []).
   difference_s([S|Ss], List2, Difference) :-
	 difference_s(Ss, List2, Ds),
	 (member_s(S, List2)
	     -> Difference = Ds
	     ;  Difference = [S|Ds]).

/*%   generalize_term(+TERM, -GENERAL_TERM)
%      GENERAL_TERM has the same function and arity as TERM
%      but all its arguments are new and different variables.
*/

  generalize_term(Term, Template) :-
     functor(Term, Functor, Arity),
     functor(Template, Functor, Arity).

/*%  member_s(+ELEMENT, +LIST)
%      The "structure" ELEMENT is a member of LIST
*/

   member_s(Structure, [X|Subs]) :-
	Structure == X
	-> true
	 ; member_s(Structure, Subs).

/*% delete_s(+LIST, +ELEMENT, -NEWLIST)
%      delete the first ELEMENT structure from LIST to create NEWLIST
*/

  delete_s([], _Element, []).
  delete_s([X|Xs], Element, NewList) :-
      (X == Element 
       -> NewList = Xs
       ; NewList = [X|Rest],
	 delete_s(Xs, Element, Rest)).

/*%   preds_in_normalform(+PREDSPEC, -PREDICATE)
%       Normal form of specifying predicate is f(....).
%       An alternate form is Functor/Arity
%       It converts PREDSPEC to a normal form PREDICATE
*/

  pred_in_normalform(F/N, Predicate) :-!,
     functor(Predicate, F, N).
  pred_in_normalform(Predicate, Predicate).

/*%% END OF UTIL.PRO %%%*/
